#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

void producer(int pipedes, int nValues);
void consumer(int pipedes);

int main(){
  
  /* Descrittori per la pipe, il primo è per la lettura, il secondo per la scrittura */
  int pipedes[2];
  /* ID del processo produttore */
  pid_t pid_prod;
  /* ID del processo consumatore */
  pid_t pid_cons;

  /* Creazione della pipe per la comunicazione */
  if (pipe(pipedes)){
    puts("Errore nella creazione della pipe\n");
    return 1;
  }
  
  /* Creazione di un processo per il produttore. 
     Fork crea un processo identico al chiamante, l'unico modo di 
     distinguere il processo padre (quello che chiama fork) dal figlio è verificare il valore restituito dalla chiamata */
  pid_prod = fork();
  if (pid_prod == -1){
    puts("Errore nella creazione del processo per il produttore\n");
    return 1;   
  }else if (pid_prod == 0){
    /* Se pid_prod vale 0 ci troviamo nel processo figlio, quello per il produttore 
       La prima cosa da fare è chiudere il descrittore in lettura della pipe, perché il produttore deve solo scriverci */
    close(pipedes[0]);
    producer(pipedes[1], 100);
    /* Terminata l'esecuzione del lavoro dobbiamo uscire per evitare che si eseguano le istruzioni che vengono dopo, si ricordi che 
       i due processi sono completamente distinti ma uguali! */
    exit(0);
  }
  /* Creazione di un processo per il consumatore. */
  pid_cons = fork();
  if (pid_cons == -1){
    puts("Errore nella creazione del processo per il consumatore\n");
    return 1;   
  }else if (pid_cons == 0){
    /* Se pid_cons vale 0 ci troviamo nel processo figlio, quello per il consumatore.
       Chiudiamo il descrittore in scrittura e chiamiamo la funzione consumatore */
    close(pipedes[1]);
    consumer(pipedes[0]);
    exit(0);
  }

  /* Al processo padre non servono più le pipe. Ogni processo ha la propria copia delle variabili e dei descrittori, quindi
     Questa operazione non ha alcun effetto sui due figli */
  close(pipedes[0]);
  close(pipedes[1]);
    
  /* Aspetta che i due processi terminino l'esecuzione */
  wait(NULL);
  wait(NULL);
  
    
  puts("Esecuzione del programma terminata\n");
  return 0;
}

/* Consumatore. Legge dati dalla pipe a ciclo continuo e li stampa. 
   Termina l'esecuzione quando incontra il carattere 0 (zero) */
void consumer(int pipedes){   
  char value;
  puts("Consumatore: esecuzione iniziata\n");
  do{
    read(pipedes, (void *)&value, sizeof(value));
    printf("Consumatore: ho letto il valore %c\n", value);
  }while(value != '0');
  puts("Consumatore: esecuzione terminata\n");
} 


/* Produttore. Genera caratteri casuali e li inserisce nela pipe a ciclo continuo. 
   Generato il numero massimo di valori inserisce 0 (zero) nel buffer per indicare al
   consumatore che l'esecuzione è terminata */
void producer(int pipedes, int nValues){
  char value;
  int i = 0;
  srand(time(NULL));
  puts("Produttore: esecuzione iniziata\n");
  do{
    if (i < nValues)
      value = (char)(65 + ((int)(rand() % 25)));
    else
      value = '0';
    write(pipedes, (void *)&value, sizeof(value));
    printf("Produttore: ho inserito il valore %c\n", value);
    i++;
  }while(i <= nValues);
  puts("Produttore: esecuzione terminata\n");
}
