#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <math.h>
#include <time.h>

#define BUFFER_SIZE 10

void *producer(void *params);
void *consumer(void *params);

/* Parametri da passare ai due thread */
struct thread_params{
  char *buffer;
  pthread_mutex_t *mutex;
  sem_t *semaphore_space;
  sem_t *semaphore_data;
  /* Questi valori non servono a entrambi i thread, ma possiamo evitare di creare due strutture */
  int *first_element;
  int *next_element;
  int nValues;
};

int main(){
  
  /* Creazione del buffer */
  char buffer[BUFFER_SIZE];
  /* Prossimo elemento da leggere e prossimo da scrivere sul buffer 
     Non creiamo questi valori nelle funzioni di produttore e consumatore perché
     potremmo avere più di un'istanza di ciascuno! */
  int first_element = 0, next_element = 0;
  
  /* Creazione ed inizializzazione dei semafori per l'utilizzo del buffer condiviso 
     Il primo è un mutex, il secondo indica lo spazio disponibile sul buffer, il terzo i dati presenti
   */
  pthread_mutex_t mutex;
  sem_t semaphore_space;
  sem_t semaphore_data;
  if (pthread_mutex_init(&mutex, NULL)){
    puts("Errore nell'inizializzazione del mutex \n");
    return 1;
  }
  if (sem_init(&semaphore_space, 0, BUFFER_SIZE) == -1){
    puts("Errore nell'inizializzazione del mutex \n");
    return 1;
  }
  if (sem_init(&semaphore_data, 0, 0) == -1){
    puts("Errore nell'inizializzazione del mutex \n");
    return 1;
  }
  
  /* ID del thread produttore */
  pthread_t tid_prod;
  /* ID del thread consumatore */
  pthread_t tid_cons;
  /* Parametri per i thread. 
     Dato che entrambi usano gli stessi valori, quindi possiamo usare la stessa struttura */
  struct thread_params params = {buffer, &mutex, &semaphore_space, &semaphore_data, &first_element, &next_element, 100};
 
  /* Creazione di un thread per il produttore. */
  if (pthread_create(&tid_prod, NULL, producer, (void*)&params)){
    puts("Errore nella creazione del thread produttore\n");
    return 1;
  }
  /* Creazione di un thread per il consumatore. */
  if (pthread_create(&tid_cons, NULL, consumer, (void*)&params)){
    puts("Errore nella creazione del thread consumatore\n");
    return 1;
  }

  /* Aspetta che i due thread terminino l'esecuzione */
  pthread_join(tid_prod, NULL);
  pthread_join(tid_cons, NULL); 
    
  puts("Esecuzione del programma terminata\n");
  return 0;
}

/* Consumatore. Legge dati dalla pipe a ciclo continuo e li stampa. 
   Termina l'esecuzione quando incontra il carattere 0 (zero) */
void *consumer(void *params){
  char value;
  /* Alias per i parametri */
  struct thread_params *par = (struct thread_params *)params;
  /* Alias per il prossimo elemento da leggere sul buffer */
  int *first_element = par->first_element;

  puts("Consumatore: esecuzione iniziata\n");
  do{
    /* INIZIO SEZIONE CRITICA */
    sem_wait(par->semaphore_data); /* Wait sul semaforo dati disponibili */
    pthread_mutex_lock(par->mutex); /* Lock del mutex del buffer condiviso */
    value = par->buffer[*first_element];
    *first_element = (*first_element + 1) % BUFFER_SIZE;
    sem_post(par->semaphore_space); /* Post sul semaforo spazio disponibile */   
    pthread_mutex_unlock(par->mutex); /* Unlock del mutex del buffer condiviso */
    /* FINE SEZIONE CRITICA */
    printf("Consumatore: ho letto il valore %c\n", value);
  }while(value != '0');
  puts("Consumatore: esecuzione terminata\n");
} 


/* Produttore. Genera caratteri casuali e li inserisce nela pipe a ciclo continuo. 
   Generato il numero massimo di valori inserisce 0 (zero) nel buffer per indicare al
   consumatore che l'esecuzione è terminata */
void *producer(void *params){
  char value;
  /* Alias per i parametri */
  struct thread_params *par = (struct thread_params *)params;
  /* Alias per il prossimo elemento da scrivere sul buffer */
  int *next_element = par->next_element;
  int i = 0;
    
  srand(time(NULL));
  puts("Produttore: esecuzione iniziata\n");
  do{
    if (i < par->nValues)
      value = (char)(65 + ((int)(rand() % 25)));
    else
      value = '0';

    /* INIZIO SEZIONE CRITICA */
    sem_wait(par->semaphore_space); /* Wait sul semaforo spazio disponibile */
    pthread_mutex_lock(par->mutex); /* Lock del mutex del buffer condiviso */
    /*Incrementa il contatore del buffer, che è trattato come una coda circolare*/
    par->buffer[*next_element] = value;
    *next_element = (*next_element + 1) % BUFFER_SIZE;
    sem_post(par->semaphore_data); /* Post sul semaforo dati disponibili */   
    pthread_mutex_unlock(par->mutex); /* Unlock del mutex del buffer condiviso */
    /* FINE SEZIONE CRITICA */

    printf("Produttore: ho inserito il valore %c\n", value);
    i++;
  }while(i <= par->nValues);
  puts("Produttore: esecuzione terminata\n");
}
